//-----------------------------------------------------------------------------
// File: distanceModulus
// Description: Fit 3-parameter models to the latest compilation of Type Ia 
//              supernova data.
//
//              a    - universal scale factor
//
//              OM   - Omega_M
//              OL   - Omega_L
//              H    - roughly related to Hubble's constant
//
//              For some models OM and OL may have different meanings.
//
// Created: June 2008 HBP
// Updated for ESHEP 2012, La Pommeraye, Anjou, France
//-----------------------------------------------------------------------------
#include <cmath>
#include <cassert>
#include <iostream>
using namespace std;
//-----------------------------------------------------------------------------
double OFFSET=5*log10(2.99*pow(10.0, 5.0))+25;
//-----------------------------------------------------------------------------

// Model with cosmological constant
double lambdaModel(double a, double OM, double OL)
{
  // a^3 * [Omega_M/a^3 + (1-Omega_M-Omega_L)/a^2 + Omega_L]
  double y = OM + (1 - OM - OL)*a + OL*a*a*a;
  if ( y > 0 )
    return y;
  else
    return 1.e50;
}

// Dungan-Prosper Varying-G or phantom energy model
double phantomModel(double a, double OM, double OL)
{
  OL = OL; // to avoid compiler warnings
  return exp(OM*(a-1));
}

// random phantom energy model
double phantomModel2(double a, double OM, double OL)
{
  return exp(OM*a+OL*a*a);
}


// pick model
double (*model)(double, double, double) = lambdaModel;

// compute distance modulus (see AJP article)
double distanceModulus(double z, double OM, double OL, double H)
{
  int N=200;
  double F = 0;
  double a = 1.0/(1+z);
  double h = (1-a) / N;
  
  for(int i=0; i < N; i++)
    {
      double x = a + (i+0.5)*h;
      F = F + 1.0/sqrt(x * model(x, OM, OL));
    }

  F = F*h;
  double OK = 1 - OM - OL;

  if ( OK != 0 )
    {      
      double rootOK = sqrt(fabs(OK));
      double theta  = rootOK * F;
      if ( OK > 0 )
        F = sinh(theta);
      else
        F = sin(theta);
      F /= rootOK;
    }

  double y = 5*log10((1+z)*F/H) + OFFSET;
  return y;
}


