#!/usr/bin/env python
#------------------------------------------------------------------
# Tutorial 3 - Fitting Type Ia supernovae data
# Created sometime in the 21st century Harrison B. Prosper
#$Revision:$
#------------------------------------------------------------------
import os,sys
from ROOT import *
from histutil import *
from time import sleep
#------------------------------------------------------------------
def makePlots(data, OM, OL, H, zmin, zmax, mumin, mumax):
	
	# Create data plot
	# Copy data into arrays
	xx = array('d')
	ex = array('d')
	yy = array('d')
	ey = array('d')
	
	ndata = data.numEntries()
	for ii in xrange(ndata):
		row = data.get(ii)
		xx.append(row['z'].getVal())
		ex.append(0)
		yy.append(row['x'].getVal())
		ey.append(row['dx'].getVal())

	g = mkgraphErrors(xx, yy, ex, ey,
			  "redshift z",
			  "distance modulus #mu", 
			  zmin, zmax,
			  ymin=mumin,
			  ymax=mumax, color=kBlack)
	g.SetMarkerSize(0.2)

	# Create model plot
	nz = 100
	zstep = (zmax-zmin)/nz
	zz = array('d')
	mu = array('d')
	omegaM = OM.getVal()
	omegaL = OL.getVal()
	q = H.getVal()
	
	for ii in xrange(nz):
		zz.append( (ii+0.5)*zstep )
		mu.append( distanceModulus(zz[-1],
					   omegaM,
					   omegaL, q) )
		
	gfit = mkgraph(zz, mu,
		       "redshift z", "distance modulus #mu", 
		       zmin, zmax, color=kRed, lwidth=2)
	return (g, gfit)
#------------------------------------------------------------------	    
def main():
	print "\n","="*80
	print "\t\tType Ia Supernovae"
	print "="*80

	# compile distance modulus function
	gROOT.ProcessLine('.L distanceModulus.cc+')
	
	# Suppress all messages except those that matter
	RooMsgService.instance().setGlobalKillBelow(RooFit.FATAL)

        #----------------------------------------------------------		
	# Make the workspace
	wspace = RooWorkspace('TypeIa')

	# define ranges
	zmin = 0.0 
	zmax = 1.5
	mumin = 32.0
	mumax = 48.0

	# create parameter objects
	name = RooStringVar('name', 'TypeIa', 'name of type Ia')
	name.Print()

	# data parameters
	# red shift
	wspace.factory('z[%f, %f]' % (zmin, zmax))
	z = wspace.var('z')

	# distance modulus
	wspace.factory('x[%f, %f]' % (mumin, mumax))
	x = wspace.var('x')
	wspace.factory('dx[0, 2]')
	dx = wspace.var('dx')

	# model parameters
	wspace.factory('OM[0.5, 0, 200]')
	OM = wspace.var('OM')
	wspace.factory('OL[0.5, 0, 200]')
	OL = wspace.var('OL')
	wspace.factory('H[70, 0, 200]')
	H = wspace.var('H')
	
        #----------------------------------------------------------
	# Read Type Ia data
	#----------------------------------------------------------
	data = RooDataSet.read('SCPUnion2.1.txt',
			       RooArgList(name, z, x, dx))
	ndata = data.numEntries()
	print "number of observations: %d" % ndata
	
	for ii in range(ndata):
		if ii % 100 == 0:
			d = data.get(ii)
			print "%5d %-8s\t%10.3f\t%10.4f +/- %-10.4f"%\
			      (ii,
			       d['name'].getVal(),
			       d['z'].getVal(),
			       d['x'].getVal(),
			       d['dx'].getVal())			
	# import data into workspace (give it a name first)
	data.SetName('data')
	getattr(wspace,'import')(data)
	
	# distance modulus
	# note use of compiled C++ program
	mu = RooFormulaVar('mu', '#mu',
			   'distanceModulus(z, OM, OL, H)',
			   RooArgList(z, OM, OL, H))
	getattr(wspace,'import')(mu)
	wspace.factory('Gaussian::model(x, mu, dx)')
	model  = wspace.pdf('model')

	# Now fit
	print "="*80
	swatch = TStopwatch()
	swatch.Start()

	# Save results of fit to object results
	results = model.fitTo(data, RooFit.Save())
	print "real time: %10.3f s" % swatch.RealTime()
	print "="*80
	results.Print()
	print "="*80

        #----------------------------------------------------------            # Plot results
        #----------------------------------------------------------
	# see histutil.py
	setStyle()
	
	g, gfit = makePlots(data, OM, OL, H, zmin, zmax, mumin, mumax)
	
	# plot
	canvas = TCanvas("fig_results", "SN1a", 10, 10, 500, 500)
	canvas.cd()
	g.Draw("ap")
	gfit.Draw("c same")
	
	wrtext = Scribe(0.25,0.45)
	wrtext.write("The Union2.1 Compilation")
	wrtext.write("The Supernova Cosmology Project")
	wrtext.write("http://supernova.lbl.gov/Union/figures/")

	canvas.SaveAs(".gif")
	canvas.SaveAs(".pdf")

	sleep(10)
#------------------------------------------------------------------
try:
	main()
except KeyboardInterrupt:
	print
	print "bye cruel world!"
	print
	


